@@ -10,18 +10,39 @@ from .sdax_core import (
     AsyncDagTaskProcessor,
     AsyncDagTaskProcessorBuilder,
     AsyncTaskProcessor,
+    SdaxExecutionError,
+    flatten_exceptions,
+)
+from .tasks import (
+    AsyncTask,
+    RetryableException,
+    SdaxTaskGroup,
+    TaskFunction,
+    join,
+    task,
+    task_func,
+    task_group_func,
+    task_group_task,
+    task_sync_func,
 )
-from .tasks import AsyncTask, RetryableException, SdaxTaskGroup, TaskFunction
 
-__version__ = "0.5.2"
+__version__ = "0.7.1"
 
 __all__ = [
     "AsyncTask",
     "RetryableException",
     "SdaxTaskGroup",
     "TaskFunction",
+    "join",
+    "task",
+    "task_func",
+    "task_group_func",
+    "task_group_task",
+    "task_sync_func",
     "AsyncTaskProcessor",
     "AsyncDagTaskProcessor",
     "AsyncDagTaskProcessorBuilder",
     "AsyncDagLevelAdapterBuilder",
+    "SdaxExecutionError",
+    "flatten_exceptions",
 ]
