@@ -3,19 +3,22 @@
 Tests complex dependency graphs to validate wave construction assumptions.
 """
 
+from typing import Any, Hashable
+
 import pytest
+
 from sdax.sdax_core import AsyncTask, TaskFunction
 from sdax.sdax_task_analyser import TaskAnalyzer
 
 
 # Helper functions for creating test tasks
-async def dummy_func(ctx):
+async def dummy_func(ctx: Any):
     """Dummy function for testing."""
     pass
 
 
 def make_task(
-    name: str,
+    name: Hashable,
     has_pre: bool = False,
     has_exec: bool = False,
     has_post: bool = False,
@@ -112,11 +115,9 @@ def format_wave_structure(graph) -> str:
     """
     lines = []
     for wave in graph.waves:
-        deps = wave.depends_on_tasks
-        deps_str = (
-            f"depends_on_tasks={deps}" if deps else "no deps"
-        )
-        tasks_str = ", ".join(wave.tasks)
+        deps = tuple(str(dep) for dep in wave.depends_on_tasks)
+        deps_str = f"depends_on_tasks={deps}" if deps else "no deps"
+        tasks_str = ", ".join(str(task) for task in wave.tasks)
         lines.append(f"Wave {wave.wave_num}: [{tasks_str}] ({deps_str})")
     return "\n".join(lines)
 
@@ -126,7 +127,7 @@ class TestBasicGraphs:
 
     def test_simple_linear_chain(self):
         """A -> B -> C: Simple linear dependency chain."""
-        analyzer = TaskAnalyzer()
+        analyzer = TaskAnalyzer[Any, str]()
         analyzer.add_task(make_task("A", has_pre=True, has_post=True), depends_on=())
         analyzer.add_task(make_task("B", has_pre=True, has_post=True), depends_on=("A",))
         analyzer.add_task(make_task("C", has_pre=True, has_post=True), depends_on=("B",))
@@ -162,6 +163,36 @@ class TestBasicGraphs:
         assert post_wave_b is not None and post_wave_b.tasks == ("B",)
         assert post_wave_a is not None and post_wave_a.tasks == ("A",)
 
+    def test_integer_identifier_chain(self):
+        """1 -> 2 -> 3: Same chain but using integer task identifiers."""
+        analyzer = TaskAnalyzer[Any, int]()
+        analyzer.add_task(make_task(1, has_pre=True, has_post=True), depends_on=())
+        analyzer.add_task(make_task(2, has_pre=True, has_post=True), depends_on=(1,))
+        analyzer.add_task(make_task(3, has_pre=True, has_post=True), depends_on=(2,))
+
+        analysis = analyzer.analyze()
+
+        wave1 = analysis.pre_wave_containing(1)
+        wave2 = analysis.pre_wave_containing(2)
+        wave3 = analysis.pre_wave_containing(3)
+
+        assert wave1 is not None and wave1.tasks == (1,)
+        assert wave1.depends_on_tasks == ()
+
+        assert wave2 is not None and wave2.tasks == (2,)
+        assert set(wave2.depends_on_tasks) == {1}
+
+        assert wave3 is not None and wave3.tasks == (3,)
+        assert set(wave3.depends_on_tasks) == {2}
+
+        post_wave1 = analysis.post_wave_containing(1)
+        post_wave2 = analysis.post_wave_containing(2)
+        post_wave3 = analysis.post_wave_containing(3)
+
+        assert post_wave3 is not None and post_wave3.tasks == (3,)
+        assert post_wave2 is not None and set(post_wave2.depends_on_tasks) == {3}
+        assert post_wave1 is not None and set(post_wave1.depends_on_tasks) == {2}
+
     def test_independent_chains(self):
         """A -> B  and  C -> D: Two independent chains.
 
@@ -671,6 +702,8 @@ class TestPostExecuteGraphs:
         assert w0.depends_on_tasks == ()
         assert w1.tasks == ("B",)
         assert set(w1.depends_on_tasks) == {"D"}
+
+
 class TestErrorCases:
     """Test error handling."""
 
@@ -735,7 +768,7 @@ class TestRandomGraph:
         """
         import random
 
-        #random.seed(42)  # For reproducibility
+        # random.seed(42)  # For reproducibility
 
         analyzer = TaskAnalyzer()
         task_names = [f"T{i:02d}" for i in range(20)]
